#ifndef ConvImpl2D_ref_cpp_H_
#define ConvImpl2D_ref_cpp_H_

#include <array>
#include <memory>
#include <tuple>
#include <vector>

#include "backend/OperatorImpl.hpp"
#include "operator/Conv.hpp"
#include "utils/Registrar.hpp"
#include "utils/Types.h"

namespace Aidge {
// class Conv_Op;

// compute kernel registry for forward and backward
class ConvImpl2DForward_ref_cpp
    : public Registrable<std::tuple<DataType, DataType, DataType, DataType>,
                         void(const Conv_Op<2>::Parameters &, const std::array<DimSize_t, 4> &, const void *,
                              const void *, const void *, void *)> {};
class ConvImpl2DBackward_ref_cpp
    : public Registrable<std::tuple<DataType, DataType, DataType, DataType>,
                         void(const Conv_Op<2>::Parameters &, const std::array<DimSize_t, 4> &, const void *,
                              const void *, const void *, void *)> {};

class ConvImpl2D_ref_cpp : public OperatorImpl {
   private:
    const Conv_Op<2> &mOp;
    std::array<NbElts_t, 3> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    ConvImpl2D_ref_cpp(const Conv_Op<2> &op) : mOp(op), mNbConsumedData({0, 0, 0}), mNbProducedData({0}) {}

    static std::unique_ptr<ConvImpl2D_ref_cpp> create(const Conv_Op<2> &op) {
        return std::make_unique<ConvImpl2D_ref_cpp>(op);
    }

   public:
    NbElts_t getNbRequiredData(IOIndex_t inputIdx) const override final;
    NbElts_t getNbRequiredProtected(IOIndex_t inputIdx) const override final;
    NbElts_t getRequiredMemory(IOIndex_t outputIdx, const std::vector<DimSize_t> &inputsSize) const override final;
    NbElts_t getNbConsumedData(IOIndex_t inputIdx) const override final;
    NbElts_t getNbProducedData(IOIndex_t outputIdx) const override final;

    void forward();

    void backward();
};

namespace {
// add ref_cpp backend to Conv_Op<2> implementation registry
static Registrar<Conv_Op<2>> registrarConvImpl2D_ref_cpp("cpu", Aidge::ConvImpl2D_ref_cpp::create);
}  // namespace
}  // namespace Aidge

#endif /* ConvImpl2D_ref_cpp_H_ */