

#ifndef ConvImpl2D_ref_cpp_forward_kernel_H_
#define ConvImpl2D_ref_cpp_forward_kernel_H_

#include "utils/Registrar.hpp"

#include "operator/ConvImpl.hpp"
#include "utils/Types.h"
#include <array>
#include <algorithm>

namespace Aidge {
/**
 * @brief Forward kernel for 2D Convolution on CPU backend.
 * @tparam I Input data type.
 * @tparam W Weight data type.
 * @tparam B Bias data type.
 * @tparam O Output data type.
 * @param params tuple of Parameters from the Operator
 * @param dims Array of input dimensions.
 * @param input_ const input Tensor.
 * @param weights_ const weight Tensor.
 * @param biases_ const Biais Tensor.
 * @param output_ Output Tensor.
 */
template <class I, class W, class B, class O>
void ConvImpl2D_ref_cpp_forward_kernel(const Conv_Op<2>::Parameters &params, const std::array<DimSize_t, 4> &dims,
                                       const void *input_, const void *weights_, const void *biases_, void *output_) {
    // FIXME: missing convolution parameters as arguments
    const I *input = static_cast<const I *>(input_);
    const W *weights = static_cast<const W *>(weights_);
    const B *biases = static_cast<const B *>(biases_);
    O *output = static_cast<O *>(output_);
/*
    // output H size
    const std::size_t oxSize =
            static_cast<std::size_t>(static_cast<float>(dims[0] - std::get<4>(params)[0] + std::get<0>(params)[0]) /
                                static_cast<float>(std::get<0>(params)[0]));
    // output W size
    const std::size_t oySize =
            static_cast<std::size_t>(static_cast<float>(dims[1] - std::get<4>(params)[1] + std::get<0>(params)[1]) /
                                static_cast<float>(std::get<0>(params)[1]));

    // TODO: kernel computation
    // output (Xout, Yout, outCh, batch)
    // input  (Xin, Yin, inCh, batch)
    // weight (kernelX, kernelY, inCh, outCh)
    // does not take Dilation parameter into account
    for (std::size_t ox = 0; ox < oxSize; ++ox) {
        for (std::size_t oy = 0; oy < oySize; ++oy) {
            const std::size_t ix = ox * std::get<0>(params)[0];
            const std::size_t iy = oy * std::get<0>(params)[1];

            for (std::size_t outCh = 0; outCh < std::get<3>(params); ++outCh) {
                const std::size_t oIndex = dims[3] * (outCh + std::get<3>(params) * (oy + oySize * ox));
                B biasVal = (biases != nullptr) ? biases[outCh] : B(0);
                for (std::size_t batch = 0; batch < dims[3]; ++batch) {
                    output[oIndex + batch] = biasVal;
                }
                for (std::size_t inCh = 0; inCh < dims[2]; ++inCh) {
                    for (std::size_t sx = 0; sx < std::get<4>(params)[0]; ++sx) {
                        for (std::size_t sy = 0; sy < std::get<4>(params)[1]; ++sy) {
                            const std::size_t wIndex =
                                    outCh + std::get<3>(params) * (inCh + dims[2] * (sy + std::get<4>(params)[1] * sx));
                            std::size_t iIndex = dims[3] * (inCh + dims[2] * ((iy + sy) + dims[1] * (ix + sx)));
                            for (std::size_t batch = 0; batch < dims[3]; ++batch) {
                                output[oIndex + batch] += weights[wIndex] * input[iIndex + batch];
                            }
                        }
                    }
                }
            }
        }
    }
*/

    
    // output H size
    const std::size_t oxSize =
            static_cast<std::size_t>(static_cast<float>(dims[2] - std::get<4>(params)[0] + std::get<0>(params)[0]) /
                                static_cast<float>(std::get<0>(params)[0]));
    // output W size
    const std::size_t oySize =
            static_cast<std::size_t>(static_cast<float>(dims[3] - std::get<4>(params)[1] + std::get<0>(params)[1]) /
                                static_cast<float>(std::get<0>(params)[1]));

    // TODO: kernel computation
    // output (batch, outCh, Xout, Yout)
    // input  (batch, inCh, Xin, Yin)
    // weight (outCh, inCh, kernelX, kernelY)
    // does not take Dilation parameter into account
    for (std::size_t batch = 0; batch < dims[0]; ++batch) {
        for (std::size_t outCh = 0; outCh < std::get<3>(params); ++outCh) {
            const std::size_t oIndex = (outCh + batch*std::get<3>(params)) * oxSize * oySize;
            B biasVal = (biases != nullptr) ? biases[outCh] : B(0);
            std::fill(output + oIndex, output+(oIndex+oxSize*oySize), biasVal);
            for (std::size_t inCh = 0; inCh < dims[1]; ++inCh) {
                const std::size_t iIndex = (inCh + batch*dims[1]) * dims[2] * dims[3];
                const std::size_t wIndex = (inCh + outCh*dims[1]) * std::get<4>(params)[0] * std::get<4>(params)[1];
                for (std::size_t ox = 0; ox < oxSize; ++ox) {
                    for (std::size_t oy = 0; oy < oySize; ++oy) {
                        const std::size_t oIndexFull = oIndex + ox*oySize + oy;
                        const std::size_t ix = ox * std::get<0>(params)[0];
                        const std::size_t iy = oy * std::get<0>(params)[1];
                        for (std::size_t sx = 0; sx < std::get<4>(params)[0]; ++sx) {
                            for (std::size_t sy = 0; sy < std::get<4>(params)[1]; ++sy) {
                                output[oIndexFull] += weights[wIndex + sx*std::get<4>(params)[1] + sy] * 
                                                        input[iIndex + (ix+sx)*dims[3] + (iy+sy)];
                            }
                        }
                    }
                }
            }
        }
    }
}

namespace {
static Registrar<ConvImpl2DForward_ref_cpp> registrarConvImpl2DForward_ref_cpp_Float32(
        {DataType::Float32, DataType::Float32, DataType::Float32, DataType::Float32},
        Aidge::ConvImpl2D_ref_cpp_forward_kernel<float, float, float, float>);
static Registrar<ConvImpl2DForward_ref_cpp> registrarConvImpl2DForward_ref_cpp_Int32(
        {DataType::Int32, DataType::Int32, DataType::Int32, DataType::Int32},
        Aidge::ConvImpl2D_ref_cpp_forward_kernel<int, int, int, int>);
static Registrar<ConvImpl2DForward_ref_cpp> registrarConvImpl2DForward_ref_cpp_Float64(
        {DataType::Float64, DataType::Float64, DataType::Float64, DataType::Float64},
        Aidge::ConvImpl2D_ref_cpp_forward_kernel<double, double, double, double>);
}  // namespace
}  // namespace Aidge

#endif /* ConvImpl2D_ref_cpp_forward_kernel_H_ */
