#ifndef FCImpl_ref_cpp_H_
#define FCImpl_ref_cpp_H_

#include "backend/OperatorImpl.hpp"
#include "operator/FC.hpp"
#include "utils/Registrar.hpp"
#include "utils/Types.h"
#include <memory>
#include <vector>
#include <array>

namespace Aidge {
// class FC_Op;

// compute kernel registry for forward and backward
class FCImplForward_ref_cpp : public Registrable<std::tuple<DataType, DataType, DataType, DataType>,
                                                 void(const FC_Op::Parameters &, const DimSize_t, const DimSize_t,
                                                      const void *, const void *, const void *, void *)> {};
class FCImplBackward_ref_cpp : public Registrable<std::tuple<DataType, DataType, DataType, DataType>,
                                                  void(const FC_Op::Parameters &, const DimSize_t, const DimSize_t,
                                                       const void *, const void *, const void *, void *)> {};

class FCImpl_ref_cpp : public OperatorImpl {
   private:
    const FC_Op &mOp;
    std::array<NbElts_t, 3> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    FCImpl_ref_cpp(const FC_Op &op) : mOp(op), mNbConsumedData({0, 0, 0}), mNbProducedData({0}) {}

    static std::unique_ptr<FCImpl_ref_cpp> create(const FC_Op &op) { return std::make_unique<FCImpl_ref_cpp>(op); }

   public:
    NbElts_t getNbRequiredData(IOIndex_t inputIdx) const override final;
    NbElts_t getNbRequiredProtected(IOIndex_t inputIdx) const override final;
    NbElts_t getRequiredMemory(IOIndex_t outputIdx, const std::vector<DimSize_t> &inputsSize) const override final;
    NbElts_t getNbConsumedData(IOIndex_t inputIdx) const override final;
    NbElts_t getNbProducedData(IOIndex_t outputIdx) const override final;

    void forward();

    void backward();
};

namespace {
static Registrar<FC_Op> registrarFCImpl_ref_cpp("cpu", Aidge::FCImpl_ref_cpp::create);
}
}  // namespace Aidge

#endif /* FCImpl_ref_cpp_H_ */