#ifndef ProducerImpl_ref_cpp_H_
#define ProducerImpl_ref_cpp_H_

#include "utils/Types.h"

#include "backend/OperatorImpl.hpp"
#include "operator/Producer.hpp"
#include "utils/Registrar.hpp"
#include <memory>

namespace Aidge {
template <DimIdx_t DIM>
class ProducerImpl_ref_cpp : public OperatorImpl {
   private:
    const Producer_Op<DIM> &mOp;

   public:
    ProducerImpl_ref_cpp(const Producer_Op<DIM> &op) : mOp(op) {}

    static std::unique_ptr<ProducerImpl_ref_cpp> create(const Producer_Op<DIM> &op) {
        return std::make_unique<ProducerImpl_ref_cpp>(op);
    }

   public:
    NbElts_t getNbRequiredData(IOIndex_t inputIdx) const override final;
    NbElts_t getNbRequiredProtected(IOIndex_t inputIdx) const override final;
    NbElts_t getRequiredMemory(IOIndex_t outputIdx, const std::vector<DimSize_t> &inputsSize) const override final;
    NbElts_t getNbConsumedData(IOIndex_t inputIdx) const override final;
    NbElts_t getNbProducedData(IOIndex_t outputIdx) const override final;

    void forward();

    void backward();
};

namespace {
static Registrar<Producer_Op<1>> registrarProducer1DImpl_ref_cpp("cpu", Aidge::ProducerImpl_ref_cpp<1>::create);
static Registrar<Producer_Op<2>> registrarProducer2DImpl_ref_cpp("cpu", Aidge::ProducerImpl_ref_cpp<2>::create);
static Registrar<Producer_Op<3>> registrarProducer3DImpl_ref_cpp("cpu", Aidge::ProducerImpl_ref_cpp<3>::create);
static Registrar<Producer_Op<4>> registrarProducer4DImpl_ref_cpp("cpu", Aidge::ProducerImpl_ref_cpp<4>::create);
static Registrar<Producer_Op<5>> registrarProducer5DImpl_ref_cpp("cpu", Aidge::ProducerImpl_ref_cpp<5>::create);
}  // namespace
}  // namespace Aidge

#endif /* ProducerImpl_ref_cpp_H_ */