#ifndef ReLUImpl_ref_cpp_H_
#define ReLUImpl_ref_cpp_H_

#include "backend/OperatorImpl.hpp"
#include "operator/ReLU.hpp"
#include "utils/Registrar.hpp"
#include <memory>
#include <vector>

namespace Aidge {
// class ReLU_Op;

// compute kernel registry for forward and backward
class ReLUImplForward_ref_cpp
    : public Registrable<std::tuple<DataType, DataType>, void(const ReLU_Op::Parameters&, std::size_t, const void*, void*)> {
};
class ReLUImplBackward_ref_cpp
    : public Registrable<std::tuple<DataType, DataType>, void(const ReLU_Op::Parameters&, std::size_t, const void*, void*)> {
};

class ReLUImpl_ref_cpp : public OperatorImpl {
   private:
    const ReLU_Op& mOp;
    std::array<NbElts_t, 1> mNbConsumedData;
    std::array<NbElts_t, 1> mNbProducedData;

   public:
    ReLUImpl_ref_cpp(const ReLU_Op& op) : mOp(op), mNbConsumedData({0}), mNbProducedData({0}) {}

    static std::unique_ptr<ReLUImpl_ref_cpp> create(const ReLU_Op& op) {
        return std::make_unique<ReLUImpl_ref_cpp>(op);
    }

   public:
    NbElts_t getNbRequiredData(IOIndex_t inputIdx) const override final;
    NbElts_t getNbRequiredProtected(IOIndex_t inputIdx) const override final;
    NbElts_t getRequiredMemory(IOIndex_t outputIdx, const std::vector<DimSize_t>& inputsSize) const override final;
    NbElts_t getNbConsumedData(IOIndex_t inputIdx) const override final;
    NbElts_t getNbProducedData(IOIndex_t outputIdx) const override final;

    void forward();

    void backward();
};

namespace {
static Registrar<ReLU_Op> registrarReLUImpl_ref_cpp("cpu", Aidge::ReLUImpl_ref_cpp::create);
}
}  // namespace Aidge

#endif /* ReLUImpl_ref_cpp_H_ */