
#include <cassert>
#include <numeric>
#include <chrono>
#include <thread>

#include "operator/ReLU.hpp"

#include "operator/ReLUImpl.hpp"
#include "operator/ReLUImpl_kernels.hpp"
#include "utils/Types.h"
#include <numeric>
#include <vector>

// FIXME: replace whole Tensor with minimum needed data quantity
Aidge::NbElts_t Aidge::ReLUImpl_ref_cpp::getNbRequiredData(Aidge::IOIndex_t /*inputIdx*/) const {
    assert(mOp.getInput(0) && "requires valid input");

    // Requires the whole tensors
    const auto& inputDims = std::static_pointer_cast<Tensor>(mOp.getInput(0))->dims();

    return std::accumulate(inputDims.begin(), inputDims.end(),
                        static_cast<NbElts_t>(1), std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::ReLUImpl_ref_cpp::getNbRequiredProtected(Aidge::IOIndex_t /*inputIdx*/) const {
    // for the direct convolution algorithm, convolutions can be in-place, if there is no padding!
    return 0;
}

Aidge::NbElts_t Aidge::ReLUImpl_ref_cpp::getRequiredMemory(Aidge::IOIndex_t /*outputIdx*/, const std::vector<Aidge::DimSize_t>& /*inputsSize*/) const {
    const auto& outputDims = std::static_pointer_cast<Tensor>(mOp.getOutput(0))->dims();
    return std::accumulate(outputDims.begin(), outputDims.end(),
                        static_cast<NbElts_t>(1), std::multiplies<NbElts_t>());
}

Aidge::NbElts_t Aidge::ReLUImpl_ref_cpp::getNbConsumedData(Aidge::IOIndex_t /*inputIdx*/) const {
    return mNbConsumedData[0];
}

Aidge::NbElts_t Aidge::ReLUImpl_ref_cpp::getNbProducedData(Aidge::IOIndex_t /*outputIdx*/) const {
    return mNbProducedData[0];
}

void Aidge::ReLUImpl_ref_cpp::forward() {
    // FIXME: uncomment the following code once memory handling will work
    assert(mOp.mInputs[0] && "missing input #0");

    // Find the correct kernel type
    auto kernelFunc = Registrar<ReLUImplForward_ref_cpp>::create({
        mOp.mInputs[0]->dataType(),
        mOp.mOutput->dataType()});

    // Call kernel
    kernelFunc(mOp.getParams(),
        std::static_pointer_cast<Tensor>(mOp.getInput(0))->size(),
        mOp.mInputs[0]->getImpl()->rawPtr(),
        mOp.mOutput->getImpl()->rawPtr());


    mNbConsumedData[0]+= getNbRequiredData(0); // each input is consumed by the minimum amount for a forward pass

    mNbProducedData[0]+= getRequiredMemory(0, {});
}

void Aidge::ReLUImpl_ref_cpp::backward() {
    printf("Not implemented yet.\n");
}
